/*
 * grfdebug.cpp
 * Reads grfdebug.log and the specified NFO file to generate a more useful
 * debug log.
 * 
 * Copyright 2007 by Dale McCoy.
 * dalestan@gmail.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include<iostream>
#include<iomanip>
#include<string>
#include<fstream>

using namespace std;

typedef unsigned int uint;

#include"nfosprite.h"
#include"allocarray.h"

void read_file(istream&in,AllocArray<Sprite>&sprites);

uint nfogrfid;

int FindReal(AllocArray<Sprite>& sprites,int num){
	uint count=0;
	while(num && ++count < sprites.size())
		if(sprites[count]->GetType()==Sprite::ST_REAL)
			num--;
	if(num)return -1;
	return count;
}

void ParseSprite(Pseudo *pSprite, int num, int&var, int&mask){
	mask=-1;
	var=-1;
	if(!pSprite)return;
	Pseudo &sprite=*pSprite;
	if(sprite.size()<7||sprite[0]!=2||!(sprite[3]&0x80)||sprite[3]==0x80||sprite[3]==0x83)
		return;

	uint w = 1<<((sprite[3]>>2)&3);
	if(w == 2) mask = 0xFFFF;
	if(w == 1) mask = 0xFF;
	uint loc=4;//Start at first <var>
	while(num--){//read <var> [<param>] <varadjust> [<op> ...]. loc reports byte to be checked.
		if((sprite[loc++/*<var>*/]&0xE0)==0x60)loc++;//<param>
		if(!(sprite[loc]&0x20))return;//end of advanced
		if(sprite[loc++/*<shift>*/]&0xC0)
			loc+=2*w;//<add> and <div>/<mod>
		loc+=w;//<and>
		loc++;//<op>
	}
	var = sprite[loc++];
	if ((var&0xE0)==0x60){
		var <<= 8;
		var |= sprite[loc];
	}
}

uint FixResult(uint rslt){
	rslt&=0x7FFF;
	if((rslt>>8)==0x7F)
		rslt&=0xFF;
	return rslt;
}

int main(int argc, char *argv[]){
	if(argc<=1){
		cout<<"No NFO file specified.\n";
		return 1;
	}
	
	ifstream fin("grfdebug.log"), nfo(argv[1]);
	string name=argv[1];
	if(!fin){
		cout<<"Could not open grfdebug.log.\n";
		return 1;
	}
	
	if(!nfo.is_open())
		nfo.open((name+".nfo").c_str());
	if(!nfo.is_open())
		nfo.open(("sprites/"+name).c_str());
	if(!nfo.is_open())
		nfo.open(("sprites/"+name+".nfo").c_str());
	if(!nfo.is_open()){
		cout<<"Could not open \""<<name<<"\" specified on the command-line.\n";
		return 1;
	}
	nfo.clear();

	AllocArray<Sprite>sprites;
	read_file(nfo,sprites);
	nfo.close();

	ofstream fout("grfdebug.txt");
	if(!fout){
		cout<<"Could not open grfdebug.txt for writing.\n";
		return 1;
	}

	fin>>setbase(16);
	fout<<setbase(16);

	string cmd;
	uint dword1, dword2;
	uint feat = 0;
	Pseudo* sprite = NULL;
	uint varcount = 0;

#define feature		(dword1&0xFF)
#define id			((dword1>>8)&0xFF)
#define callback	(dword1>>16)

	while(fin){
		fin>>cmd>>dword1>>dword2;
		if(cmd == "GET"){
			feat=feature;
			if(callback == 1)
				fout<<"Getting random triggers";
			else if(callback)
				fout<<"Calling callback "<<callback;
			else
				fout<<"Retrieving graphics";
			fout<<" for feature "<<feature<<", ID "<<id;
			if(dword2 == nfogrfid)
				fout<<".\n";
			else
				fout<<". (WARNING! GRFid given in grfdebug.log does not match GRFid given in NFO file.)\n";
		}else if(cmd == "ACT2"){
			fout<<"Processing sprite "<<setbase(10)<<dword1+1<<setbase(16);
			if(sprites.size()<=dword1+1 ||
					sprites[dword1+1]->GetType() != Sprite::ST_PSEUDO ||
					(*(Pseudo*)(sprites[dword1+1]))[0] != 2){
				fout<<".  (WARNING! Sprite in NFO file is not an action 2.)\n";
				sprite = NULL;
			}else{
				sprite = (Pseudo*)sprites[dword1+1];
				varcount=0;
				fout<<".\n";
			}
		}else if(cmd == "RND")
			fout<<"Testing bits "<<dword2<<" of random value "<<dword1<<".\n";
		else if(cmd == "VAR"){
			int var,widthmask;
			ParseSprite(sprite, varcount++, var, widthmask);
			fout<<"Testing variable";
			if(var>=0){
				fout<<' ';
				if(var&0xFF00)
					fout<<(var>>8)<<", param "<<(var&0xFF);
				else
					fout<<var;
			}
			fout<<". Shifted: "<<dword1;
			if(dword1!=dword2)
				fout<<", Adjusted: "<<dword2;
			if(dword2!=(dword2&widthmask))
				fout<<", Masked: "<<(dword2&widthmask);
			fout<<".\n";
		}else if(cmd == "RSLT")
			if(dword1&0x10000)
				fout<<"Returning calculated callback result "<<FixResult(dword1)<<".\n";
			else
				fout<<"Returning callback result "<<FixResult(dword1)<<".\n";
		else if(cmd == "SPRT"){
			uint base[] = {0x1324,0x4000, 0x6CDC, 0x99B8, 0xBBCD, 0xDDE2}, i = 0;
			while(dword1>base[i+1])i++;
			fout<<"Returning real sprite number "<<setbase(10)<<FindReal(sprites,dword1-base[i])<<setbase(16)<<".\n";
		}else
			fout<<"Unknown line: "<<cmd<<" "<<setw(8)<<dword1<<" "<<dword2<<setw(8)<<".\n";
	}
}
